"""
This module provides the Status class, which encapsulates
a status code for Nagios.
"""

class Status(object):
    """
    Encapsulates a Nagios status, which holds a name and
    an exit code.
    """

    def __init__(self, name, exit_code):
        """
        Creates a new status object for Nagios with the given name and
        exit code.

        **Note**: In general, this should never be called since the standard
        statuses are exported from ``pynagios``.
        """
        self.name = name
        self.exit_code = exit_code

    def __repr__(self):
        return "Status(name=%s, exit_code=%d)" % (repr(self.name), self.exit_code)



"""
Tools for creating performance data for Nagios plugin responses.
If you're adding performance data to a :py:class:`~pynagios.response.Response`
object, then :py:func:`~pynagios.response.Response.set_perf_data` can be
called instead of having to create an entire :py:class:`PerfData` object.
"""

import re
from range import Range

class PerfData(object):
    """
    This class represents performance data for a response. Since
    performance data has a non-trivial response format, this class
    is meant to ease the formation of performance data.
    """

    def __init__(self, label, value, uom=None, warn=None, crit=None,
                 minval=None, maxval=None):
        """Creates a new object representing a single performance data
        item for a Nagios response.

        Performance data is extra key/value data that can be returned
        along with a response. The performance data is not used immediately
        by Nagios itself, but can be extracted by 3rd party tools and can
        often be helpful additional information for system administrators
        to view. The `label` can be any string, but `value` must be a
        numeric value.

        Raises :class:`ValueError` if any of the parameters are invalid.
        The exact nature of the error is in the human readable message
        attribute of the exception.

        :Parameters:
          - `label`: Label for the performance data. This must be a
            string.
          - `value`: Value of the data point. This must be a number whose
            characters are in the class of `[-0-9.]`
          - `uom` (optional): Unit of measure. This must only be `%`, `s`
            for seconds, `c` for continous data, or a unit of bit space
            measurement ('b', 'kb', etc.)
          - `warn` (optional): Warning range for this metric.
          - `crit` (optional): Critical range for this metric.
          - `minval` (optional): Minimum value possible for this metric,
            if one exists.
          - `maxval` (optional): Maximum value possible for this metric,
            if one exists.
        """
        self.label = label
        self.value = value
        self.uom = uom
        self.warn = warn
        self.crit = crit
        self.minval = minval
        self.maxval = maxval

    @property
    def value(self):
        """The value of this metric."""
        return self._value

    @value.setter
    def value(self, value):
        if value is None:
            raise ValueError("value must not be None")
        elif not self._is_valid_value(value):
            raise ValueError("value must be in class [-0-9.]")

        self._value = value

    @property
    def warn(self):
        """
        The warning range of this metric. This return value of this
        will always be a :py:class:`~pynagios.range.Range` object, even
        if it was set with a string.
        """
        return self._warn

    @warn.setter
    def warn(self, value):
        if value is not None and not isinstance(value, Range):
            value = Range(value)

        self._warn = value

    @property
    def crit(self):
        """
        The critical range of this metric. This return value of this
        will always be a :py:class:`~pynagios.range.Range` object,
        even if it was set with a string.
        """
        return self._crit

    @crit.setter
    def crit(self, value):
        if value is not None and not isinstance(value, Range):
            value = Range(value)

        self._crit = value

    @property
    def minval(self):
        """
        The minimum value possible for this metric. This doesn't make
        a lot of sense if the `uom` is '%', since that is obviously going
        to be 0, but this will return whatever was set.
        """
        return self._minval

    @minval.setter
    def minval(self, value):
        if not self._is_valid_value(value):
            raise ValueError("minval must be in class [-0-9.]")

        self._minval = value

    @property
    def maxval(self):
        """
        The maximum value possible for this metric. This doesn't make
        a lot of sense if the `uom` is '%', since that is obviously going
        to be 100, but this will return whatever was set.
        """
        return self._maxval

    @maxval.setter
    def maxval(self, value):
        if not self._is_valid_value(value):
            raise ValueError("maxval must be in class [-0-9.]")

        self._maxval = value

    @property
    def uom(self):
        """
        The unit of measure (UOM) for this metric.
        """
        return self._uom

    @uom.setter
    def uom(self, value):
        valids = ['', 's', '%', 'b', 'kb', 'mb', 'gb', 'tb', 'c']
        if value is not None and not str(value).lower() in valids:
            raise ValueError("uom must be in: %s" % valids)

        self._uom = value

    def __str__(self):
        """
        Returns the proper string format that should be outputted
        in the plugin response string. This format is documented in
        depth in the Nagios developer guidelines, but in general looks
        like this:

            | 'label'=value[UOM];[warn];[crit];[min];[max]

        """
        # Quotify the label
        label = self._quote_if_needed(self.label)

        # Check for None in each and make it empty string if so
        uom = self.uom or ''
        warn = self.warn or ''
        crit = self.crit or ''
        minval = self.minval or ''
        maxval = self.maxval or ''

        # Create the proper format and return it
        return "%s=%s%s;%s;%s;%s;%s" % (label, self.value, uom, warn, crit, minval, maxval)

    def _is_valid_value(self, value):
        """
        Returns boolean noting whether a value is in the proper value
        format which certain values for the performance data must adhere to.
        """
        value_format = re.compile(r"[-0-9.]+$")
        return value is None or value_format.match(str(value))

    def _quote_if_needed(self, value):
        """
        This handles single quoting the label if necessary. The reason that
        this is not done all the time is so that characters can be saved
        since Nagios only reads 80 characters and one line of stdout.
        """
        if '=' in value or ' ' in value or "'" in value:
            # Quote the string and replace single quotes with double single
            # quotes and return that
            return "'%s'" % value.replace("'", "''")
        else:
            return value




"""
This module provides the Plugin class, which is the basic
class which encapsulates a single plugin. This is the class
which should be subclassed when creating new plugins.
"""

import sys
from copy import copy
from optparse import Option, OptionParser, make_option
from range import Range, RangeValueError
from response import Response
from status import Status

# Status constants which contain both the status text and the
# exit status associated with them.
OK = Status("OK", 0)
WARNING = Status("WARN", 1)
CRITICAL = Status("CRIT", 2)
UNKNOWN = Status("UNKNOWN", 3)

def check_pynagios_range(option, opt, value):
    """
    This parses and returns the Nagios range value.
    """
    try:
        return Range(value)
    except RangeValueError, e:
        raise OptionValueError("options %s: %s" % (opt, e.message))

# Add the "pynagios_range" type to the global available
# options for OptionParser
Option.TYPES = Option.TYPES + ("pynagios_range",)
Option.TYPE_CHECKER["pynagios_range"] = check_pynagios_range

class PluginMeta(type):
    """
    We use a metaclass to create the plugins in order to gather and
    set up things such as command line arguments.
    """

    def __new__(cls, name, bases, attrs):
        attrs = attrs if attrs else {}

        # Set the options on the plugin by finding all the Options and
        # setting them. This also removes the original Option attributes.
        options = []

        for key,val in attrs.items():
            if isinstance(val, Option):
                # We set the destination of the Option to always be the
                # attribute key...
                val.dest = key

                # Append it to the list of options and delete it from
                # the original attributes list
                options.append(val)
                del attrs[key]

        # Need to iterate through the bases in order to extract the
        # list of parent options, so we can inherit those.
        for base in bases:
            if hasattr(base, "_options"):
                options.extend(getattr(base, "_options"))

        # Store the option list and create the option parser
        attrs["_options"] = options
        attrs["_option_parser"] = OptionParser(option_list=options)

        # Create the class
        return super(PluginMeta, cls).__new__(cls, name, bases, attrs)

class Plugin(object):
    """
    Encapsulates a single plugin. This is able to parse the command line
    arguments, understands the range syntax, provides help output, and
    more.
    """
    __metaclass__ = PluginMeta

    hostname = make_option("-H", "--hostname", type="string", default=None)
    warning = make_option("-w", "--warning", type="pynagios_range")
    critical = make_option("-c", "--critical", type="pynagios_range")
    timeout = make_option("-t", "--timeout", type="int", default=0)
    verbosity = make_option("-v", "--verbose", action="count")

    # TODO: Still missing version

    def __init__(self, args=sys.argv):
        """
        Instantiates a plugin, setting up the options and arguments state.
        Initialization by itself shouldn't do much, since the plugin should run
        when :py:func:`check` is called.

        This init method will parse the arguments given in ``args`` and will
        set the results on the ``options`` attribute. If no ``args`` are given,
        the command line arguments given to the whole Python application will
        be used.

        All plugins parse standard command line arguments that are required
        by the Nagios developer guidelines:

          - ``hostname`` - Set via ``-H`` or ``--hostname``, this should be the
            host that this check targets, if applicable.
          - ``warning`` - Set via ``-w`` or ``--warning``, this should be a valid
            range in which the value of the plugin is considered to be a warning.
          - ``critical`` - Set via ``-c`` or ``--critical``, this should be a
            valid range in which the value is considered to be critical.
          - ``timeout`` - Set via ``-t`` or ``--timeout``, this is an int value
            for the timeout of this check.
          - ``verbosity`` - Set via ``-v``, where additional ``v`` means more
            verbosity. Example: ``-vvv`` will set ``options.verbosity`` to 3.

        Subclasses can define additional options by creating ``Option`` instances
        and assigning them to class attributes. The easiest way to make an
        ``Option`` is to use Python's built-in ``optparse`` methods. The following
        is an example plugin which adds a simple string argument:::

            class MyPlugin(Plugin):
                your_name = make_option("--your-name", dest="your_name", type="string")

        Instantiating the above plugin will result in the value of the new
        argument being available in ``options.your_name``.
        """
        # Parse the given arguments to set the options
        (self.options, self.args) = self._option_parser.parse_args(args)

    def check(self):
        """
        This method is what should be called to run this plugin and return
        a proper :py:class:`~pynagios.response.Response` object. Subclasses
        are expected to implement this.
        """
        raise NotImplementedError("This method must be implemented by the plugin.")

    def response_for_value(self, value, message=None):
        """
        This method is meant to be used by plugin implementers to return a
        valid :py:class:`~pynagios.response.Response` object for the given value.
        The status of this response is determined based on the warning and
        critical ranges given via the command line, which the plugin automatically
        parses.

        An optional ``message`` argument may be provided to set the message
        for the Response object. Note that this can easily be added later as well
        by simply setting the message attribute on the response object returned.

        Creating a response using this method from :py:func:`check` makes it
        trivial to calculate the value, grab a response, set some performance
        metrics, and return it.
        """
        status = OK
        if self.options.critical is not None and self.options.critical.in_range(value):
            status = CRITICAL
        elif self.options.warning is not None and self.options.warning.in_range(value):
            status = WARNING

        return Response(status, message=message)



"""
Contains a class to represent a range that adheres to the range
format defined by Nagios.
"""

class RangeValueError(ValueError):
    """
    This exception is raised when an invalid value is passed to
    :py:class:`Range`. The message of this exception will contain a
    human readable explanation of the error.
    """
    pass

class Range(object):
    """
    Encapsulates a Nagios range value. This is the value which would
    be passed to command line arguments involving ranges. Examples
    of ranges:

      10 - < 0 OR > 10
      10:20 - < 10 OR > 20
      @10:20 - >= 10 AND <= 20
      ~:10 - > 10 (~ = -inf)
      10:~ - < 10

    """

    def __init__(self, value):
        """
        Initializes a Nagios range with the given value. The value should be
        in the Nagios range format, which is the following:

        ::

          [@]start:end

        Notes:

          - ``start`` must be less than or equal to ``end``
          - Ranges by default are exclusive. A range of ``10:20`` will match values
            that are ``< 10 OR >20``.
          - ``@`` means the range is `inclusive`. So ``@10:20`` is valid in the
            case that the value is ``>= 10 AND <= 20``.
          - If ``start`` or ``end`` is ``~``, this value is negative or positive
            inifinity, respectively. A range of ``~:20`` will match values that
            are ``> 20`` only.
          - If ``start`` is not given, then it is assumed to be 0.
          - If ``end`` is not given, but a ``:`` exists, then ``end`` is assumed
            to be infinity. Example: ``5:`` would match ``< 5``.
        """
        # Clean up the value by clearing whitespace. Also note that an empty
        # value is invalid.
        value = value.strip()
        if len(value) == 0: raise RangeValueError('Range must not be empty')

        # Test for the inclusivity marked by the '@' symbol at the beginning
        if value.startswith('@'):
            self.inclusive = True
            value = value[1:]
        else:
            self.inclusive = False

        # Split by the ':' character to get the start/end parts.
        parts = value.split(':')
        if len(parts) == 1: parts.insert(0, '0')

        # Parse the start value. If no ':' is included in value (e.g. '10')
        # then the start is assumed to be 0. Otherwise, it is an integer
        # value which can possibly be infinity.
        try:
            if parts[0] == '~':
                start = float("-inf")
            else:
                start = float(parts[0])
        except ValueError:
            raise RangeValueError('invalid start value: %s' % parts[0])

        # Parse the end value, which can be positive infinity.
        try:
            if parts[1] == '' or parts[1] == '~':
                end = float("inf")
            else:
                end = float(parts[1])
        except ValueError:
            raise RangeValueError('invalid end value: %s' % parts[1])

        # The start must be less than the end
        if start > end: raise RangeValueError('start must be less than or equal to end')

        # Set the instance variables
        self.start = start
        self.end = end

    def in_range(self, value):
        """
        Tests whether ``value`` is in this range.
        """
        if self.inclusive:
            return value >= self.start and value <= self.end
        else:
            return value < self.start or value > self.end

    def __str__(self):
        """
        Turns this range object back into a valid range string which can
        be passed to another plugin or used for debug output. The string returned
        from here should generally be equivalent to the value given to the
        constructor, but sometimes it can be slightly different. However,
        it will always be functionally equivalent.

        Examples:

        ::

          >> str(Range("@10:20")) == "@10:20"
          >> str(Range("10")) == "~:10"
          >> str(Range("10:")) == "10:~"
        """
        result = '@' if self.inclusive else ''

        # Setup some proxy variables since typing `self` all the time is tiring
        start = self.start
        end = self.end

        # Only put start in the result if it is not equal to 0, since
        # it can otherwise be omitted
        if start == float("-inf"):
            result += '~:'
        elif start != 0.0:
            value = int(start) if start.is_integer() else start
            result += str(value) + ':'

        # Append the end value next
        if end == float("inf"):
            result += '~'
        else:
            value = int(end) if end.is_integer() else end
            result += str(value)

        return result



"""
Contains the class which represents a response for Nagios. This
encapsulates the response format that Nagios expects.
"""

import sys
from perf_data import PerfData

class Response(object):
    """
    This class represents a response from a Nagios plugin. Nagios plugins
    must respond in a very specific format, and this plugin assists by
    providing helpers which make emitting this format simple.
    """

    def __init__(self, status=None, message=None):
        """
        Nagios responses are expected to be in a very specific format, and
        this class allows these responses to easily be built up and extracted
        in the proper format.

        This class makes it easy to set the status, message, and performance
        data for a response.

        :Parameters:
          - `status` (optional): A :py:class:`~pynagios.status.Status` object
            representing the status of the response.
          - `message` (optional): An information message to include with the
            output.
        """
        self.status = status
        self.message = message
        self.perf_data = {}

    def set_perf_data(self, label, value, uom=None, warn=None, crit=None,
                      minval=None, maxval=None):
        """
        Adds performance data to the response. Performance data is shown
        in the Nagios GUI and can be used by 3rd party programs to build
        graphs or other informational output. There are many options to this
        method. They are the same as the initialization parameters for a
        :py:class:`~pynagios.perf_data.PerfData` object.

        .. seealso:: :py:class:`~pynagios.perf_data.PerfData`
        """
        # Just set the perf data on the dictionary. PerfData handles
        # argument validation.
        self.perf_data[label] = PerfData(label, value, uom=uom, warn=warn,
                                         crit=crit, minval=minval,
                                         maxval=maxval)

    def exit(self):
        """
        This prints out the response to ``stdout`` and exits with the
        proper exit code.
        """
        print(str(self))
        sys.exit(self.status.exit_code)

    def __str__(self):
        """
        The string format of this object is the valid Nagios output
        format. The response format is expected to be the following:

        ::

          status: information|performance data

        An example of realistic output:

        ::

          OK: 27 users logged in|users=27;0:40;0:60;0;
        """
        result = "%s:" % self.status.name

        if self.message is not None:
            result += " %s" % self.message

        if len(self.perf_data) > 0:
            # Attach the performance data to the result
            data = [str(val) for key,val in self.perf_data.iteritems()]
            result += '|%s' % (' '.join(data))

        return result



"""
This package provides all the modules for writing a Nagios plugin
with Python. The package file itself exports the constants used
throughout the library.
"""

from optparse import make_option

from plugin import Plugin, OK, WARNING, CRITICAL, UNKNOWN
from range import Range
from response import Response
from status import Status

version = '0.1.1'
"""Current version of PyNagios"""



from setuptools import setup, Command

class PyTest(Command):
    """
    This builds a subcommand for setup.py which allows tests to be
    run on the library without installing it. Tests can be run with:

        python setup.py test

    """
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        import sys, subprocess
        errno = subprocess.call([sys.executable, 'test.py'])
        raise SystemExit(errno)

# Get the long description by reading the README
try:
    readme_content = open("README.rst").read()
except:
    readme_content = ""

# Create the actual setup method
setup(name='pynagios',
      version='0.1.1',
      description='Python library to write Nagios plugins.',
      long_description=readme_content,
      author='Mitchell Hashimoto',
      author_email='mitchell@kiip.me',
      maintainer='Mitchell Hashimoto',
      maintainer_email='mitchell@kiip.me',
      url="https://kiip.github.com/pynagios/",
      license="MIT License",
      keywords=["nagios", "pynagios", "monitoring"],
      packages=['pynagios'],
      cmdclass= { 'test': PyTest },
      classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: System :: Systems Administration"]
      )
